<?php

namespace Drupal\entity_custom_base_field\Form;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Field\FieldStorageDefinitionListenerInterface;
use Drupal\Core\Form\ConfirmFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Url;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Defines a confirmation form to import base field.
 */
class ConfirmBaseFieldImportForm extends ConfirmFormBase {

  /**
   * Entity type of the base field to delete.
   *
   * @var string
   */
  protected $entityType;

  /**
   * Machine name of the base field to delete.
   *
   * @var string
   */
  protected $fieldName;

  /**
   * Entity Field Manager service.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * Messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * Config factory service.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * Logger factory service.
   *
   * @var \Drupal\Core\Logger\LoggerChannelFactoryInterface
   */
  protected $loggerFactory;

  /**
   * Field storage definition listener service.
   *
   * @var \Drupal\Core\Field\FieldStorageDefinitionListenerInterface
   */
  protected $fieldStorageDefnListener;

  /**
   * Constructs a \Drupal\entity_custom_base_field\ConfirmDeleteBaseFieldForm.
   *
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entityFieldManager
   *   Entity Field Manager.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   Messenger.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $configFactory
   *   Config factory.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $loggerFactory
   *   Logger factory.
   * @param \Drupal\Core\Field\FieldStorageDefinitionListenerInterface $fieldStorageDefnListener
   *   Field Storage definition Listener.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   Entity Type Manager.
   */
  public function __construct(EntityFieldManagerInterface $entityFieldManager, MessengerInterface $messenger, ConfigFactoryInterface $configFactory, LoggerChannelFactoryInterface $loggerFactory, FieldStorageDefinitionListenerInterface $fieldStorageDefnListener, EntityTypeManagerInterface $entityTypeManager) {
    $this->entityFieldManager = $entityFieldManager;
    $this->messenger = $messenger;
    $this->configFactory = $configFactory;
    $this->loggerFactory = $loggerFactory;
    $this->fieldStorageDefnListener = $fieldStorageDefnListener;
    $this->entityTypeManager = $entityTypeManager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    // Instantiates this form class.
    return new static(
          // Load the service required to construct this class.
          $container->get('entity_field.manager'),
          $container->get('messenger'),
          $container->get('config.factory'),
          $container->get('logger.factory'),
          $container->get('field_storage_definition.listener'),
          $container->get('entity_type.manager'),
      );
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    // Get the base fields from config.
    $baseFieldsConfig = $this->configFactory->get('entity_custom_base_field.settings')->get('data');

    foreach ($baseFieldsConfig as $fieldConfig) {
      // Get the base field definition which needs to be deleted.
      try {
        $this->entityTypeManager->clearCachedDefinitions();
        $this->entityFieldManager->clearCachedFieldDefinitions();
        $definition = $this->entityFieldManager->getFieldStorageDefinitions($fieldConfig['entity_types'])[$fieldConfig['field_machine_name']] ?? NULL;
        if (isset($definition)) {
          $this->fieldStorageDefnListener->onFieldStorageDefinitionCreate($definition);
        }
        $this->messenger->addStatus($this->t("Successfully imported all the base fields."));
      }
      catch (\Exception $e) {
        // Show warning message.
        $this->messenger->addWarning($this->t("There was an error while trying to import the base field %field for entity type %type. Please check the recent log messages for more information.",
        [
          '%field' => $fieldConfig['field_machine_name'],
          '%type' => $fieldConfig['entity_types'],
        ]));
        // Log the details.
        $this->loggerFactory->get('entity_custom_base_field')->error($this->t("The following error occurred while trying to import the base field %field for entity type %type : %message",
        [
          '%field' => $fieldConfig['field_machine_name'],
          '%type' => $fieldConfig['entity_types'],
          '%message' => $e->getMessage(),
        ]));
      }
    }

    // Redirect to Custom base field config page.
    $form_state->setRedirect('entity_custom_base_field.settings');
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() : string {
    return "confirm_delete_base_field_form";
  }

  /**
   * {@inheritdoc}
   */
  public function getCancelUrl() {
    return new Url('entity_custom_base_field.settings');
  }

  /**
   * {@inheritdoc}
   */
  public function getQuestion() {
    return $this->t('Do you want to create all the base fields from the <em>config</em> into <em>database</em>?</br> Please note that this will create <em>new columns</em> in the related base tables for the entity types.');
  }

}
