<?php

namespace Drupal\entity_custom_base_field\Form;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityDefinitionUpdateManagerInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Form\ConfirmFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Plugin\CachedDiscoveryClearerInterface;
use Drupal\Core\Url;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Defines a confirmation form for deletion of a custom base entity field.
 */
class ConfirmDeleteBaseFieldForm extends ConfirmFormBase {

  /**
   * Entity type of the base field to delete.
   *
   * @var string
   */
  protected $entityType;

  /**
   * Machine name of the base field to delete.
   *
   * @var string
   */
  protected $fieldName;

  /**
   * Entity Field Manager service.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * Entity Definition Update Manager service.
   *
   * @var \Drupal\Core\Entity\EntityDefinitionUpdateManagerInterface
   */
  protected $entityDefnManager;

  /**
   * Plugin cache clearer service.
   *
   * @var \Drupal\Core\Plugin\CachedDiscoveryClearerInterface
   */
  protected $pluginCacheClearer;

  /**
   * Messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * Config factory service.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * Logger factory service.
   *
   * @var \Drupal\Core\Logger\LoggerChannelFactoryInterface
   */
  protected $loggerFactory;

  /**
   * Constructs a \Drupal\entity_custom_base_field\ConfirmDeleteBaseFieldForm.
   *
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entityFieldManager
   *   Entity Field Manager.
   * @param \Drupal\Core\Entity\EntityDefinitionUpdateManagerInterface $entityDefnManager
   *   Entity Definition Update Manager.
   * @param \Drupal\Core\Plugin\CachedDiscoveryClearerInterface $pluginCacheClearer
   *   Plugin cache clearer.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   Messenger.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $configFactory
   *   Config factory.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $loggerFactory
   *   Logger factory.
   */
  public function __construct(EntityFieldManagerInterface $entityFieldManager, EntityDefinitionUpdateManagerInterface $entityDefnManager, CachedDiscoveryClearerInterface $pluginCacheClearer, MessengerInterface $messenger, ConfigFactoryInterface $configFactory, LoggerChannelFactoryInterface $loggerFactory) {
    $this->entityFieldManager = $entityFieldManager;
    $this->entityDefnManager = $entityDefnManager;
    $this->pluginCacheClearer = $pluginCacheClearer;
    $this->messenger = $messenger;
    $this->configFactory = $configFactory;
    $this->loggerFactory = $loggerFactory;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    // Instantiates this form class.
    return new static(
          // Load the service required to construct this class.
          $container->get('entity_field.manager'),
          $container->get('entity.definition_update_manager'),
          $container->get('plugin.cache_clearer'),
          $container->get('messenger'),
          $container->get('config.factory'),
          $container->get('logger.factory'),
      );
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state, string $entity_type = NULL, string $field_name = NULL) {
    $this->entityType = $entity_type;
    $this->fieldName = $field_name;
    return parent::buildForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    // Get the base field definition which needs to be deleted.
    $definition = $this->entityFieldManager->getFieldStorageDefinitions($this->entityType)[$this->fieldName] ?? NULL;

    if (isset($definition)) {

      try {
        // Uninstall field from definition & delete from database table.
        $this->entityDefnManager->uninstallFieldStorageDefinition($definition);
        // Clear the plugins cache.
        $this->pluginCacheClearer->clearCachedDefinitions();
        // Get existing config data for entity base fields.
        $configData = $this->configFactory->get('entity_custom_base_field.settings')->get('data');
        // Delete the base field data from entity_custom_base_field.settings.
        unset($configData[$this->entityType . '_' . $this->fieldName]);
        // Save the updated config.
        $this->configFactory->getEditable('entity_custom_base_field.settings')
          ->set('data', $configData)
          ->save();
        $this->messenger->addStatus($this->t("Successfully deleted the base field %field for entity type %type",
        [
          '%field' => $this->fieldName,
          '%type' => $this->entityType,
        ]));
      }
      catch (\Exception $e) {
        // Show warning message.
        $this->messenger->addWarning($this->t("There was an error while trying to delete the base field %field for entity type %type. Please check the recent log messages for more information",
        [
          '%field' => $this->fieldName,
          '%type' => $this->entityType,
        ]));
        // Log the details.
        $this->loggerFactory->get('entity_custom_base_field')->error($this->t("The following error occurred while trying to delete the base field %field for entity type %type : %message",
        [
          '%field' => $this->fieldName,
          '%type' => $this->entityType,
          '%message' => $e->getMessage(),
        ]));
      }
    }
    else {
      $this->messenger->addWarning($this->t("Unable to find any base field definition for field %field for entity type %type",
      [
        '%field' => $this->fieldName,
        '%type' => $this->entityType,
      ]));
    }

    // Redirect to Custom base field config page.
    $form_state->setRedirect('entity_custom_base_field.settings');
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() : string {
    return "confirm_delete_base_field_form";
  }

  /**
   * {@inheritdoc}
   */
  public function getCancelUrl() {
    return new Url('entity_custom_base_field.settings');
  }

  /**
   * {@inheritdoc}
   */
  public function getQuestion() {
    return $this->t('Do you want to delete the base field %field from the entity type %type? Please note that this will delete all data stored for the base field in database as well.',
    [
      '%field' => $this->fieldName,
      '%type' => $this->entityType,
    ]);
  }

}
