<?php

namespace Drupal\entity_custom_base_field\Form;

use Drupal\Core\Cache\CacheTagsInvalidatorInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Field\FieldStorageDefinitionListenerInterface;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Plugin\CachedDiscoveryClearerInterface;
use Drupal\Core\Url;
use Drupal\Core\Utility\LinkGeneratorInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Configure entity_custom_base_field settings for this site.
 */
class CustomBaseFieldSettingsForm extends ConfigFormBase {

  /**
   * Config settings.
   *
   * @var string
   */
  const SETTINGS = 'entity_custom_base_field.settings';

  /**
   * Entity Type Manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Entity Field Manager service.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * Cache tags invalidator service.
   *
   * @var \Drupal\Core\Cache\CacheTagsInvalidatorInterface
   */
  protected $cacheTagsInvalidator;

  /**
   * Module handler service.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * Field storage definition listener service.
   *
   * @var \Drupal\Core\Field\FieldStorageDefinitionListenerInterface
   */
  protected $fieldStorageDefnListener;

  /**
   * Messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * Link generator service.
   *
   * @var \Drupal\Core\Utility\LinkGeneratorInterface
   */
  protected $linkGenerator;

  /**
   * Plugin cache clearer service.
   *
   * @var \Drupal\Core\Plugin\CachedDiscoveryClearerInterface
   */
  protected $pluginCacheClearer;

  /**
   * Logger factory service.
   *
   * @var \Drupal\Core\Logger\LoggerChannelFactoryInterface
   */
  protected $loggerFactory;

  /**
   * Constructs a \Drupal\entity_custom_base_field\CustomBaseFieldSettingsForm.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   Entity Type Manager.
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entityFieldManager
   *   Entity Field Manager.
   * @param \Drupal\Core\Cache\CacheTagsInvalidatorInterface $cacheTagsInvalidator
   *   Cache tage invalidator.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $moduleHandler
   *   Module handler.
   * @param \Drupal\Core\Field\FieldStorageDefinitionListenerInterface $fieldStorageDefnListener
   *   Field Storage definition Listener.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   Messenger.
   * @param \Drupal\Core\Utility\LinkGeneratorInterface $linkGenerator
   *   Link generator.
   * @param \Drupal\Core\Plugin\CachedDiscoveryClearerInterface $pluginCacheClearer
   *   Plugin cache clearer.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $loggerFactory
   *   Logger factory.
   */
  public function __construct(EntityTypeManagerInterface $entityTypeManager, EntityFieldManagerInterface $entityFieldManager, CacheTagsInvalidatorInterface $cacheTagsInvalidator, ModuleHandlerInterface $moduleHandler, FieldStorageDefinitionListenerInterface $fieldStorageDefnListener, MessengerInterface $messenger, LinkGeneratorInterface $linkGenerator, CachedDiscoveryClearerInterface $pluginCacheClearer, LoggerChannelFactoryInterface $loggerFactory) {
    $this->entityTypeManager = $entityTypeManager;
    $this->entityFieldManager = $entityFieldManager;
    $this->cacheTagsInvalidator = $cacheTagsInvalidator;
    $this->moduleHandler = $moduleHandler;
    $this->fieldStorageDefnListener = $fieldStorageDefnListener;
    $this->messenger = $messenger;
    $this->linkGenerator = $linkGenerator;
    $this->pluginCacheClearer = $pluginCacheClearer;
    $this->loggerFactory = $loggerFactory;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    // Instantiates this form class.
    return new static(
          // Load the service required to construct this class.
          $container->get('entity_type.manager'),
          $container->get('entity_field.manager'),
          $container->get('cache_tags.invalidator'),
          $container->get('module_handler'),
          $container->get('field_storage_definition.listener'),
          $container->get('messenger'),
          $container->get('link_generator'),
          $container->get('plugin.cache_clearer'),
          $container->get('logger.factory'),
      );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'entity_custom_base_field_settings';
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames() {
    return [
      static::SETTINGS,
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {

    $form = parent::buildForm($form, $form_state);

    $config = $this->config(static::SETTINGS);

    $existingFields = $config->get('data');

    // Create a table of fields which have been added previously.
    // Table header.
    $header = [
      'entity_type' => $this->t('Entity Type'),
      'field_type' => $this->t('Field Type'),
      'machine_name' => $this->t('Machine Name'),
      'field_label' => $this->t('Label'),
      'operation' => $this->t('Operation'),
    ];
    $rows = [];
    $caption = '';
    $entityTypes = [
      'block_content' => $this->t('Content Block'),
      'node' => $this->t('Content'),
      'taxonomy_term' => $this->t('Taxonomy Term'),
      'user' => $this->t('User'),
    ];
    $fieldTypeList = [
      'boolean' => 'Boolean',
      'string' => 'Text (plain)',
      'string_long' => 'Text (plain, long)',
      'text' => 'Text (formatted)',
      'text_long' => 'Text (formatted, long)',
      'text_with_summary' => 'Text (formatted, long, with summary)',
    ];
    // Flag variable to check if all fields have been imported from config.
    $fieldExistsInDb = FALSE;

    if (!empty($existingFields)) {

      // Prepare a table of all custom base for entity types stored in config.
      $caption = $this->t("List of custom base fields created :</br>");

      foreach ($existingFields as $fieldData) {
        $rows[] = [
          'entity_type' => [
            'data' => [
              '#markup' => $entityTypes[$fieldData['entity_types']],
            ],
          ],
          'field_type' => [
            'data' => [
              '#markup' => $fieldTypeList[$fieldData['field_type']],
            ],
          ],
          'machine_name' => [
            'data' => [
              '#markup' => $fieldData['field_machine_name'],
            ],
          ],
          'field_label' => [
            'data' => [
              '#markup' => $fieldData['field_label'],
            ],
          ],
          'operation' => [
            'data' => [
              '#markup' => $this->linkGenerator->generate($this->t('Delete'), Url::fromRoute('entity_custom_base_field.delete',
              [
                'entity_type' => $fieldData['entity_types'],
                'field_name' => $fieldData['field_machine_name'],
              ]))->getGeneratedLink(),
            ],
          ],
        ];
      }

      // Check if these fields already exist in database.
      $keys = array_keys($existingFields);
      foreach ($keys as $keyVal) {
        $entityTypeKey = $existingFields[$keyVal]['entity_types'];
        // dd($entityTypeKey);
        if (!empty($entityTypeKey)) {
          $fieldDefnForEntity = $this->entityFieldManager
            ->getFieldStorageDefinitions($entityTypeKey) ?? [];
          // Field exists in database,
          // Which means base fields have been already created.
          if (!empty($fieldDefnForEntity) &&
                isset($fieldDefnForEntity[$existingFields[$keyVal]['field_machine_name']])
            ) {
            $fieldExistsInDb = TRUE;
            break;
          }
        }
      }

      // This means fields are yet to be imported into DB.
      if (!$fieldExistsInDb) {
        $caption = $this->t("List of custom base fields in config :</br>");

        // Remove delete links from Operation column.
        // As field hasn't been created in DB so far.
        foreach ($rows as &$row) {
          unset($row['operation']);
        }
        $rows[0]['operation'] = [
          'data' => [
            '#markup' => $this->t('Click on <em>') . $this->linkGenerator->generate($this->t('Import'), Url::fromRoute('entity_custom_base_field.import'))->getGeneratedLink() . $this->t('</em> to create the base fields'),
          ],
          'rowspan' => count($rows),
        ];

      }
    }

    // Now set the table element.
    $form['custom_base_fields_tabel'] = [
      '#type' => 'table',
      '#header' => $header,
      '#rows' => $rows,
      '#empty' => $this->t('No field has been created.'),
      '#caption' => $caption,
    ];

    // Allow user to create new field from UI
    // If no config has been created previously.
    // Or if all fields from config has been imported to DB.
    if ($fieldExistsInDb || empty($existingFields)) {
      $form['entity_types'] = [
        '#type' => 'select',
        '#title' => $this->t('Entity Type'),
        '#description' => $this->t('Select an entity type for which a new base field needs to be added.'),
        '#default_value' => [],
        '#options' => $entityTypes,
        '#required' => TRUE,
      ];
      $form['field_type'] = [
        '#type' => 'select',
        '#title' => $this->t('Field Type'),
        '#description' => $this->t('Select the field type viz. boolean, textarea which will be used as the type of the field being created.'),
        '#default_value' => $config->get('field_type'),
        '#options' => $fieldTypeList,
        '#required' => TRUE,
      ];

      $form['field_machine_name'] = [
        '#type' => 'machine_name',
        '#title' => $this->t('Machine Name'),
        '#maxlength' => 64,
        '#description' => $this->t('Add a unique machine name which will be used to create the base field. It must only contain lowercase letters, numbers, and underscores.'),
        '#machine_name' => [
          'exists' => [
            $this,
            'exists',
          ],
        ],
      ];

      $form['field_label'] = [
        '#type' => 'textfield',
        '#title' => $this->t('Field Label'),
        '#description' => $this->t('Add a label text which will be used for the field.'),
        '#default_value' => $config->get('field_label'),
        '#required' => TRUE,
      ];

      $form['field_description'] = [
        '#type' => 'textfield',
        '#title' => $this->t('Field Description'),
        '#description' => $this->t('Add a description text which will be used for the field.'),
        '#default_value' => $config->get('field_description'),
        '#required' => TRUE,
      ];
    }
    else {
      // Prevent user from creating new field from UI if import is pending.
      $form['actions']['submit']['#access'] = FALSE;
    }
    $form['actions']['submit']['#value'] = $this->t('Submit Data');
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    // Validate tha.
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    // Get existing config data.
    $existingFieldDetails = $this->config(static::SETTINGS)->get('data') ?? [];
    $fieldDefn = [];
    // Prepare new field config data.
    $fieldDetails[$form_state->getValue('entity_types') . '_' . $form_state->getValue('field_machine_name')] = [
      'entity_types' => $form_state->getValue('entity_types'),
      'field_type' => $form_state->getValue('field_type'),
      'field_machine_name' => $form_state->getValue('field_machine_name'),
      'field_label' => $form_state->getValue('field_label'),
      'field_description' => $form_state->getValue('field_description'),
    ];
    // Save the configuration data.
    $this->config(static::SETTINGS)
      ->set('data', array_merge($existingFieldDetails, $fieldDetails))
      ->save();

    // Prepare to invoke hook_entity_base_field_info
    // To add our new custom field definition as base field.
    $entityTypes = $this->entityTypeManager->getDefinitions();
    // Get the entity type config for which we need to add the new base field.
    $entityTypeConfig = $entityTypes[$form_state->getValue('entity_types')];

    try {
      $this->moduleHandler->invokeAllWith(
            'entity_base_field_info',
            function (callable $hook, string $module) use (&$entityTypeConfig) {
                $hook($entityTypeConfig);
            }
        );

      // Clear cache after adding new base field info for the entity.
      $this->entityTypeManager->clearCachedDefinitions();
      $this->entityFieldManager->clearCachedFieldDefinitions();
      // Get the new base field definition.
      $fieldDefn = $this->entityFieldManager->getFieldStorageDefinitions($form_state->getValue('entity_types'))[$form_state->getValue('field_machine_name')] ?? [];
      if (!empty($fieldDefn)) {
        // Add field to the entity type.
        try {
          $this->fieldStorageDefnListener->onFieldStorageDefinitionCreate($fieldDefn);
          $this->messenger->addStatus('New base field has been created successfully!!');
        }
        catch (\Exception $e) {
          // Show warning message.
          $this->messenger->addWarning($this->t("There was an error while trying to add the base field %field for entity type %type. Please check the recent log messages for more information.",
          [
            '%field' => $form_state->getValue('field_machine_name'),
            '%type' => $form_state->getValue('entity_types'),
          ]));
          // Log the details.
          $this->loggerFactory->get('entity_custom_base_field')->error($this->t("The following error occurred while trying to create the base field %field for entity type %type with error %message",
          [
            '%field' => $form_state->getValue('field_machine_name'),
            '%type' => $form_state->getValue('entity_types'),
            '%message' => $e->getMessage(),
          ]));
          // Revert current form data from config
          // And save previous data to config.
          $this->config(static::SETTINGS)
            ->set('data', $existingFieldDetails)
            ->save();
        }
      }
      else {
        // Show warning message.
        $this->messenger->addWarning($this->t("There was an error while trying to add the base field %field for entity type %type. No base field definition found.",
        [
          '%field' => $form_state->getValue('field_machine_name'),
          '%type' => $form_state->getValue('entity_types'),
        ]));
        // Revert current form data from config & save previous data to config.
        $this->config(static::SETTINGS)
          ->set('data', $existingFieldDetails)
          ->save();
      }

    }
    catch (\Exception $e) {
      // Show warning message.
      $this->messenger->addWarning($this->t("There was an error while trying to add the base field %field for entity type %type. Please check the recent log messages for more information.",
      [
        '%field' => $form_state->getValue('field_machine_name'),
        '%type' => $form_state->getValue('entity_types'),
      ]));
      // Log the details.
      $this->loggerFactory->get('entity_custom_base_field')->error($this->t("The following error occurred while trying to create the base field %field for entity type %type : %message",
      [
        '%field' => $form_state->getValue('field_machine_name'),
        '%type' => $form_state->getValue('entity_types'),
        '%message' => $e->getMessage(),
      ]));
      // Revert the current form data from config
      // and save previous data to config.
      $this->config(static::SETTINGS)
        ->set('data', $existingFieldDetails)
        ->save();
    }

    parent::submitForm($form, $form_state);
  }

  /**
   * Callback for validating machine name of field.
   *
   * @param string $submitted_value
   *   The current submitted value of the field.
   * @param array $element
   *   The element array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state object.
   *
   * @return void
   *   Returns void.
   */
  public function exists(string $submitted_value, array $element, FormStateInterface $form_state): void {
    $entityType = $form_state->getValue('entity_types');
    // Get the new base field definition.
    $fieldDefn = $this->entityFieldManager->getFieldStorageDefinitions($entityType) ?? [];

    // Check if same machine name exists in list of fields for the entity type.
    if (!empty($fieldDefn) && array_key_exists($submitted_value, $fieldDefn)) {
      // If it does exist, stop user from trying to use the same name.
      $form_state->setErrorByName('field_machine_name', $this->t("The entered machine name %name is already in use as a field name for the %type entity type. Please set a different name.",
      [
        '%name' => $submitted_value,
        '%type' => $entityType,
      ]));
    }
  }

}
